/*
Oryginalny kod: Piotr Pawluczuk / piopawlu.prv.pl
Zmiany od wersji 1.1: Rafa Lindemann / www.stamina.eu.org
*/

#include "stdafx.h"

sConnection_base::sConnection_base() {
	connectionsCount++;
	UpdateKTStatus();
	flags.callback = flags.active = flags.nocancel = flags.canceled = flags.success = flags.closed = flags.fromRequest = 0;
	this->cntID = -1;
	this->wnd = 0;
	this->next = 0;
	this->stage = esNone;
	Stage(esCreated);
	EnterCriticalSection(&CSslots);
	slots.push_back(this);
	LeaveCriticalSection(&CSslots);
	direction = edUnknown;
	operation = eoUnknown;
}
sConnection_base::~sConnection_base() {
	if (stage != esFinished)
		Stage(esFinished);
	if (wnd) {
		IMDEBUG(DBG_MISC , "[%x] Queue DestroyWindowAPC" , this);
		QueueUserAPC((PAPCFUNC)DestroyWindowAPC , hMainThread , (ULONG_PTR)this);
		while (wnd) { // Trzeba czeka, eby nie wywali drugiemu wtkowi obiektu!
			SleepEx(5,true);
			Ctrl->WMProcess();
		}
	}
	IMDEBUG(DBG_MISC , "[%x] Removing slot" , this);
	EnterCriticalSection(&CSslots);
	slots_it i = find(slots.begin() , slots.end() , this);
	if (i!=slots.end())
		slots.erase(i);
	LeaveCriticalSection(&CSslots);
	connectionsCount--;
	UpdateKTStatus();
	IMDEBUG(DBG_MISC , "[%x] Destroyed" , this);
}

int sConnection_base::CountNext() {
	int c = 0;
	sConnection_base * iter = this;
	while ((iter = iter->next)) {
		c++;
	}
	return c;
}

void sConnection_base::DoNext() {
	if (!next) return;
	if (!flags.success) {
		if (MessageBox(this->wnd , "Ostatnia operacja si nie powioda. Czy kontynuowa wysyanie pozostaych?" , "kTransfer" , MB_YESNO | MB_ICONQUESTION)==IDNO)
			return;
	}
	IMLOG("[%x] Przeczam na nastpne zapytanie" , this);
	next->wnd = this->wnd;
	// Przedstawiamy si okienku :)
	SetProp(next->wnd , "sConn*" , (HANDLE)next);
	this->wnd = 0;
	next->Stage(esCreated);
	next->Start();
	next = 0;
}

void sConnection_base::Cancel() {
	IMLOG("[%x] Anuluj" , this);
	if (next) {
		next->Cancel();
		next = 0;
	}
	if (wnd) {
		Stage(esCancel);
	}
	delete this;
}

void sConnection::init() {
	cancelEvent = CreateEvent(0 , false , false , 0);
	event = file = 0;
	sock = 0;
	protocol = 0;
	options = done = transfered = total = 0;
	thread = 0;
	startTime = 0;
	msgID = 0;
	notifyID = 0;
	UID = "";
	net = 0;
	lastUpdate = 0;
}


sConnection::sConnection() {
	init();
}

sConnection::~sConnection() {
	if (direction == edIncoming)
		incomingCount --;
	Stage(esFinished);
	DoNext();
	if (msgID) {
		sMESSAGESELECT ms;
		ms.id = msgID;
		ICMessage(IMC_MESSAGEREMOVE , (int)&ms , 1);
	}
	if (notifyID) {
		IMessage(KNotify::IM::hide , KNotify::net , IMT_ALL , notifyID);
	}
	if (thread) CloseHandle(thread);
	if (event) WSACloseEvent(event);
	if (file) CloseHandle(file);
	CloseHandle(cancelEvent);
//	CLOSESOCK(s_accept);
}

sConnection_base & sConnection_base::operator = (sConnection_base & old) {
	// "Przepinamy" potrzebne nam informacje z requesta i go usuwamy...
	this->next = old.next;
	old.next = 0;
	if (this->wnd) DestroyWindow(this->wnd);
	this->wnd = old.wnd;
	SetProp(old.wnd , "sConn*" , (HANDLE)this);
	old.wnd = 0;
	this->filename = old.filename;
	old.filename = "";
	old.flags.closed = 1;
	this->cntID = old.cntID;
	return *this;
}

sConnection & sConnection::operator = (sConnection_request & req) {
	// "Przepinamy" potrzebne nam informacje z requesta i go usuwamy...
	static_cast<sConnection_base &>(*this) = static_cast<sConnection_base &>(req);
	this->flags.fromRequest = 1;
	delete &req;
	return *this;
}

sConnection & sConnection::operator = (sConnection & req) {
	// "Przepinamy" potrzebne nam informacje z requesta i go usuwamy...
	this->port = req.port;
	this->address = req.address;
	this->UID = req.UID;
	this->direction = req.direction;
	this->operation = req.operation;
	this->net = req.net;
	static_cast<sConnection_base &>(*this) = static_cast<sConnection_base &>(req);
	return *this;
}
sConnection_request::sConnection_request(sConnection & req) {
	static_cast<sConnection_base &>(*this) = static_cast<sConnection_base &>(req);
}
sConnection::sConnection(sConnection & req) {
	init();
	*this = req;
}


void sConnection::Cancel() {
	if (stage == esFinished) return;
	this->flags.canceled = true;
	IMLOG("[%x] Anuluj" , this);
	if (next) {
		next->Cancel();
		next = 0;
	}
	
	if (wnd) {
		Stage(esCancel);
	}
	// Najlepszym sposobem na anulowanie jest wywalenie socketa... wtedy
	// cao powinna zakoczy si sama...
	if(this->sock && this->sock != INVALID_SOCKET)
	{
//		BOOL nol = TRUE;
		BOOL nol = FALSE;
		setsockopt(this->sock,SOL_SOCKET,SO_DONTLINGER,(const char*)&nol,4);
		LINGER lin = {0};
		lin.l_linger = 0;
		lin.l_onoff = 1;
		setsockopt(this->sock,SOL_SOCKET,SO_LINGER,(const char*)&lin,4);
		closesocket(this->sock);
		this->sock = 0;
	}
	// reszta wydarzy si sama...
	
}


void sConnection::Start() {
	if (direction == edIncoming)
		incomingCount ++;

	if (direction == edIncoming)
		IMLOG("[%x] Nowe poczenie przychodzce (count=%d, address=%x:%d)" , this , incomingCount , this->address , this->port);
	else
		IMLOG("[%x] Nowe poczenie wychodzce (address=%x:%d)" , this , this->address , this->port);


	// w zalenoci od kierunku i ustawie 
	thread = (HANDLE) //_beginthreadex(NULL,NULL,(OLD_THREAD)ClientThread,(void*)conn,CREATE_SUSPENDED,&conn->u_id);
		Ctrl->BeginThread(NULL,NULL,(OLD_THREAD)ClientThread,(void*)this,CREATE_SUSPENDED,&this->threadID);
	if (!thread) throw eKT("Could not create client thread!");

	if(ResumeThread(thread)==(DWORD)-1)
	{
		TerminateThread(thread,0); // I tak nie dziaa :)
		CloseHandle(thread);
		delete this;
		throw eKT("Could not start client thread!");
	}
}

void sConnection::Retry() {
	sConnection_base * conn;
	if (flags.fromRequest) 
		conn = new sConnection_request (*this);
	else
		conn = static_cast<sConnection_base*>( new sConnection (*this) );
	conn->Stage(esCreated); // niech sobi odwiey okienko...
	conn->Start();
	// skasuje si samo...
}



LONG lread = 0;
unsigned int sConnection::RecvEx(unsigned int toread,int reqall)
{
	LONG read = 0;
	LONG temp = 0;
	DWORD wr =  0;
	DWORD timeout = 60*1000;
	WSANETWORKEVENTS nev = {0};

	if(reqall > 2)
	{
		timeout = reqall;
	}
Wait:
	wr = WaitForSingleObject(this->event,timeout);
	if(wr!=WAIT_OBJECT_0)
	{
		return (reqall)?FALSE:read;
	}

	WSAEnumNetworkEvents(this->sock,this->event,&nev);
	if((nev.lNetworkEvents & FD_READ) || reqall == 2)
	{
		temp = recv(this->sock,(char*)(this->buffer + read),toread - read,0);
		if(temp < 0)
		{
			if(WSAGetLastError() == 10035)
			{
				Sleep(100);
				goto Wait;
			}
			else
			{
				return (reqall)?FALSE:read;
			}
		}
		else if(temp == 0)
		{
			return (reqall)?FALSE:read;
		}
		else
		{
			read += temp;
			lread = read;

			if(read == toread || reqall == FALSE)
				return read;
			else
			{
				Sleep(100);
				goto Wait;
			}
		}
	}
	else if(nev.lNetworkEvents & FD_CLOSE)
	{
		return (reqall)?FALSE:read;
	}
	else
		return FALSE;

}
long sConnection::SendEx(void* buffer,long length)
{
	LONG sent = 0;
	LONG temp = 0;

Send:
	temp = send(this->sock,(const char*)(buffer) + sent,length - sent,0);
	if(temp < 0)
	{
		if(GetLastError() == 10035)
		{
			Sleep(100);
			goto Send;
		}
		else return FALSE;
	}
	else if(temp == 0)
	{
		return FALSE;
	}
	else
	{
		sent += temp;
		if(sent < length)
		{
			Sleep(100);
			goto Send;
		}
		else return length;
	}
}


long sConnection::SendFile()
{
	CHAR  tbuff[256]={0};
	BYTE* buffer = this->buffer;
	DWORD length = 0;
	DWORD read = 0;
	DWORD toread = 0;
	DWORD ltc = 0;
	DWORD stc = 0;
	BOOL  last = FALSE;
	p2p_gimme_file*	gimm = (p2p_gimme_file*)buffer;
	p2p_file_hdr	file_hdr = {0};
	p2p_file*		init_file = (p2p_file*)buffer;
	this->file = CreateFile(this->filename.c_str(),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL);
	if(this->file == (HANDLE)-1)
	{
		this->ReportError("Bd otwierania pliku!",false);
		return 0;
	}
	length = GetFileSize(this->file,NULL);

	//fill struct
	memset(init_file,0,sizeof(p2p_file));
	strcpy(init_file->file_name,this->filename.substr(this->filename.find_last_of('\\')+1).c_str());
	init_file->have_file_info = 0x0003;
	this->total = init_file->size_ldw = length;
	GetFileTime(this->file,(LPFILETIME)init_file->ctime,(LPFILETIME)init_file->atime,(LPFILETIME)init_file->mtime);

	this->Stage(esNegotiating);

	this->SetStatus("Wysyam informacje o pliku...");

	DWORD code = 0x0001;
	if(!this->SendEx(&code,4))
	{
		this->ReportError("Bd wysyania!",true);
		return 0;
	}
	if(!this->SendEx(init_file,sizeof(p2p_file)))
	{
		this->ReportError("Bd wysyania!",true);
		return 0;
	}
	this->SetStatus("Oczekuj na akceptacje...");

	if(!this->RecvEx(sizeof(p2p_gimme_file),1))
	{
		this->ReportError("Bd odbierania!",true);
		return 0;
	}

	if(gimm->type != 0x0006 || gimm->start >= (int)this->total)
	{
		this->ReportError("Potwierdzenie nieprawidowe!",true);
		return 0;
	}

	this->done = gimm->start;

	if(this->done)
	{
		SetFilePointer(this->file,this->done,0,FILE_BEGIN);
	}

	this->Stage(esOperating);

	this->SetStatus("Wysyam Plik...");

	ltc = GetTickCount();

	while(this->done < this->total)
	{
		if(this->total - this->done >= 4096)
		{
			toread = 4096;
			last = ((this->total - this->done) == 0);
		}
		else
		{
			toread = this->total - this->done;
			last = TRUE;
		}
		file_hdr.chunk_size = toread;
		file_hdr.type = (last)?0x0002:0x0003;

		if(!ReadFile(this->file,buffer,toread,&read,NULL) || read != toread)
		{
			this->ReportError("Bd odczytywania pliku!",true);
			return 0;
		}

		if(!this->SendEx(&file_hdr,12))
		{
			this->ReportError("Bd wysyania nagwka!",true);
			return 0;
		}

		if(!this->SendEx(buffer,toread))
		{
			this->ReportError("Bd wysyania pakietu!",true);
			return 0;
		}

		///////////////////////////
		//calculations

		this->done += toread;
		this->transfered += toread;

		stc = GetTickCount();
		ltc = stc - ltc;
		this->UpdateInfo(ltc > 500 ? (toread / ltc) * 1000 : 0);
		ltc = stc;

		//Sleep(50);
		if(last)break;
	}


	this->SetStatus("Plik zosta wysany!");
	this->Stage(esSuccess);

	CloseHandle(this->file);
	this->file = NULL;

	return 1;
}

long sConnection::ReceiveFile()
{
	CHAR  tbuff[512]={0};
	CHAR  tname[MAX_PATH]={0};
	DWORD length = 0;
	DWORD ltc = 0;
	DWORD stc = 0;
	DWORD written = 0;
	BOOL  last = FALSE;
	DWORD chunks = 0;
	UINT64 cdate = 0;
	BYTE* buffer = this->buffer;


	p2p_gimme_file	gimm = {0};
	p2p_file*		init_file = (p2p_file*)buffer;
	p2p_file_hdr*	file_hdr = (p2p_file_hdr*)buffer;

	if(!this->RecvEx(4,120000) || *(DWORD*)buffer != 0x0001)
	{
		this->ReportError("Bd odbierania lub bdny kod operacji!",true);
		return 0;
	}

	this->SetStatus("Odbieram informacje o pliku...");

	if(!this->RecvEx(sizeof(p2p_file),120000))
	{
		this->ReportError("Bd odbierania!",true);
		return 0;
	}

	if(init_file->have_file_info != 0x0003 ||
		lstrlen(init_file->file_name)<1 || init_file->size_ldw == 0 ||
		init_file->size_hdw!=0)
	{
		this->ReportError("Odebrana informacja jest bdna, lub plik wikszy ni 4GB!",true);
		return 0;
	}


	this->filename = init_file->file_name;
	// usuwamy potencjalnie niebezpieczne znaki...
	this->filename.erase(0 , this->filename.find_last_of('\\')+1);
	this->filename.erase(0 , this->filename.find_last_of('/')+1);
	this->total = init_file->size_ldw;
	this->Stage(esNegotiating);
	this->SetStatus("Wybieram plik...");
	string dir;
	bool dirCnt = false; // katalog kontaktu...
	bool autoAccepted = false;
	ICMessage(IMC_RESTORECURDIR);
	if ((GETINT(Cfg::autoAccept) == Cfg::aaAlways)
		|| (GETINT(Cfg::autoAccept)==Cfg::aaFromAllowed && GETCNTI(cntID , Cnt::autoAccept) )) {
		autoAccepted = true;

		dir = GETCNTC(cntID , Cnt::autoDropDir);
		if (dir.size() < 2) dir = GETSTR(Cfg::autoDropDir);
		else if (dir.substr(0,2) == ".\\") {
			dirCnt = true;
			dir = GETSTR(Cfg::autoDropDir) + dir.substr(1);
		}
		char dirNew [MAX_PATH];
		ExpandEnvironmentStrings(dir.c_str() , dirNew , MAX_PATH);
		dir = dirNew;
		GetFullPathName(dir.c_str() , MAX_PATH ,  dirNew , 0);
		dir = dirNew;

		__int64 bytesAvailable;
		__int64 other;
		GetDiskFreeSpaceEx(dir.c_str() , (PULARGE_INTEGER)&bytesAvailable , (PULARGE_INTEGER)&other , (PULARGE_INTEGER)&other);
		if (this->total > 0 && (bytesAvailable < GETINT(Cfg::minDriveSpace)*1024*1024 + this->total)) {
			IMLOG("DiskFreeSpace bytesAvail=%d fileTotal=%d minSpace=%d", bytesAvailable, this->total, GETINT(Cfg::minDriveSpace));
			this->ReportError("Za mao miejsca na dysku!",true);
			return 0;
		}

		mkdirs(dir.c_str());
		char * ext = strrchr(init_file->file_name , '.');
		if (ext) {
			*(ext++) = 0;
		} else ext = "";
		int i = 0;
		do {
			if (i++)
				_snprintf(tname, MAX_PATH, "%s\\%s[%d].%s" ,dir.c_str(),init_file->file_name , i , ext);
			else
				_snprintf(tname, MAX_PATH, "%s\\%s.%s" ,dir.c_str(),init_file->file_name , ext);
		} while (!_access(tname , 0));
	} else {
		dir = GETSTR(Cfg::dirReceive);
		// potrzebujemy otwartego okna!!!
		while (!IsWindowVisible(this->wnd) && !this->flags.closed) {
			SleepEx(500 , true);
		}
		OPENFILENAME of;
		memset(&of , 0 , sizeof(of));
		of.lpstrInitialDir = dir.c_str();
		sprintf(tbuff , "Gdzie zapisa plik %s [%.2f KB]?" , this->filename.c_str() , (float)this->total / 1204);
		of.lpstrTitle = tbuff;
		of.hwndOwner = this->wnd;
		of.lStructSize=OPENFILENAME_SIZE_VERSION_400;
		char * ext = strrchr(tname , '.');
		char extBuff [50];
		if (ext) {
			_snprintf(extBuff , 50 , "Pliki *%s\0*%s\0" , ext , ext);
			of.lpstrFilter="*.*\0*.*\0";
		} else
			of.lpstrFilter="*.*\0*.*\0";
		strcpy(tname , this->filename.c_str());
		of.lpstrFile=tname;
		of.nMaxFile=MAX_PATH;
	//  of.nFilterIndex=1;
		of.Flags=OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_NOCHANGEDIR | OFN_EXPLORER;
		of.lpstrDefExt="";
		if (!GetSaveFileName(&of)) {
			this->ReportError("Anulowane przez uytkownika",false);
			return 0;
		}
		if (!autoAccepted) {
			char * dir = strdup(of.lpstrFile);
			dir[of.nFileOffset] = 0;
			SETSTR(Cfg::dirReceive , dir);
			free(dir);
		}
		__int64 bytesAvailable;
		__int64 other;
		dir = tname;
		dir.erase(dir.find_last_of('\\')+1);
		GetDiskFreeSpaceEx(dir.c_str() , (PULARGE_INTEGER)&bytesAvailable , (PULARGE_INTEGER)&other , (PULARGE_INTEGER)&other);
		if (bytesAvailable < this->total) {
			if (MessageBox(this->wnd , "Nie ma wystarczajcej iloci miejsca na dysku! Kontynuowa?" , "kTransfer" , MB_ICONQUESTION | MB_YESNO)==IDNO) {
				this->Cancel();
				return 0;
			}
		}
	}
	this->filename = tname;
	// dorzucamy resume
	sprintf(tname + strlen(tname) , ".%d.kTransfer" , init_file->size_ldw ^ init_file->mtime[0]);
	ICMessage(IMC_RESTORECURDIR);
	if(FileExists(tname,&length,&cdate) && length)
	{
		if(length < (DWORD)init_file->size_ldw && GETINT(Cfg::autoResume))
		{
			this->file = CreateFile(tname,GENERIC_WRITE,FILE_SHARE_READ,0,OPEN_EXISTING,NULL,NULL);
			this->done = length;
		}
		else
		{
			this->file = CreateFile(tname,GENERIC_WRITE,FILE_SHARE_READ,0,CREATE_ALWAYS,NULL,NULL);
			this->done = 0;
		}
	}
	else
	{
		this->file = CreateFile(tname,GENERIC_WRITE,FILE_SHARE_READ,0,CREATE_ALWAYS,NULL,NULL);
		this->done = 0;
	}

	if(this->file == (HANDLE)-1)
	{
		this->ReportError("Bd tworzenia pliku!",true);
		return 0;
	}

	try {

		if(!SetFileTime(this->file,(const FILETIME*)init_file->ctime,(const FILETIME*)init_file->atime,(const FILETIME*)init_file->mtime)) {
			throw eKT("Bd operacji I/O!",true);
		}
		this->SetStatus("Wysyam potwierdzenie pliku...");
		SetFilePointer(this->file,this->done,NULL,FILE_BEGIN);

		this->flags.active = 1;
		/////////////////
		gimm.type = 0x0006;
		gimm.dunno = 0;
		gimm.start = this->done;
		if(!this->SendEx(&gimm,12))	{
			throw eKT("Bd wysyania!",true);
		}
		//odbieranie
		this->SetStatus("Odbieram plik...");
		this->Stage(esOperating);

		ltc = GetTickCount();
		__int64 minDriveSpace = GETINT(Cfg::minDriveSpace)*1024*1024;
		while(1)
		{
			if(!this->RecvEx(12,2)) {
				throw eKT("Bd odbierania nagwka!",true);
			}

			if(file_hdr->chunk_size > 16384) {
				throw eKT("Pakiet ma wicej ni 16kB!",true);
			}
			chunks = file_hdr->chunk_size;
			last = (file_hdr->type == 2);

			if(!this->RecvEx(chunks,2)) {
				throw eKT("Bd odbierania pakietu!",true);
			}
			if(!WriteFile(this->file,this->buffer,chunks,&written,NULL) || written != chunks) {
				throw eKT("Bd zapisywania pakietu!",true);
			}

			this->done += chunks;
			this->transfered += chunks;

			stc = GetTickCount();
			ltc = stc - ltc;
			this->UpdateInfo(ltc > 500 ? (chunks / ltc) * (1000) : 0);
			__int64 bytesAvailable;
			__int64 other;
			GetDiskFreeSpaceEx(dir.c_str() , (PULARGE_INTEGER)&bytesAvailable , (PULARGE_INTEGER)&other , (PULARGE_INTEGER)&other);
			if (bytesAvailable < minDriveSpace) {
				IMLOG("DiskFreeExceeded bytesAvail=%d minSpace=%d", bytesAvailable, this->total, GETINT(Cfg::minDriveSpace));
				this->ReportError("Skoczyo si miejsce na dysku!",true);
				break;
			}

			ltc = stc;
			//Sleep(50);
			if((this->done >= this->total) || last)break;
		}
		this->flags.nocancel = true;
		this->SetStatus("Zapisuj plik...");
		CloseHandle(this->file);
		this->file = NULL;

		if(!_access(this->filename.c_str() , 0)) {
			unlink(this->filename.c_str());
		}

		if(rename(tname , this->filename.c_str())) {
			throw eKT("Bd przenoszenia pliku!",true);
		}
		this->Stage(esSuccess);
		this->SetStatus("Plik zosta odebrany!");
	} catch (eKT e) {
		ReportError(e.msg , e.severe);
		string txt = "Czy zostawi plik do docignicia pniej?\n" + this->filename;
		if (done == 0 || MessageBox(wnd , txt.c_str() , "kTransfer" , MB_YESNO | MB_ICONQUESTION)==IDNO)
			unlink(tname);
		return 0;
	}
	return 1;
}

unsigned __stdcall sConnection::ClientThread(sConnection* con)
{
	DWORD ec = 0;
	DWORD wr = 0;
	DWORD code = 0;
	LONG  read = 0;
	BOOL  result = 0;
	LINGER lin = {0};
	LONG  option = 1;
	CHAR  tbuff[256]={0};
	BYTE* buffer = con->buffer;
	try {
		con->Stage(esConnecting);
		// Nawizywanie poczenia
		if(con->direction == edOutgoing)
		{
			con->SetStatus("czenie...");
			sockaddr_in svr = {0};

			svr.sin_addr.s_addr = con->address;
			svr.sin_port = htons(con->port);
			svr.sin_family = AF_INET;


			con->flags.nocancel = true;
			con->sock = WSASocket(AF_INET,SOCK_STREAM,IPPROTO_IP,NULL,NULL,NULL);
			if(con->sock == INVALID_SOCKET)
			{
				con->ReportError("Could not create socket!",true);
				goto End;
			}
			con->flags.nocancel = false;
			if(connect(con->sock,(const sockaddr*)&svr,sizeof(svr))==-1)
			{
				con->ReportError("Could not connect!",true);
				goto End;
			}

			con->SetStatus("Poczony...");
		} // Outgoing
		//set socket options
		setsockopt(con->sock,SOL_SOCKET,SO_KEEPALIVE,(const char*)&option,sizeof(option));
/*		lin.l_linger = 2 * 1000;
		lin.l_onoff = 1;
		setsockopt(con->sock,SOL_SOCKET,SO_LINGER,(const char*)&lin,4);
*/
		option = 60*1000;
		setsockopt(con->sock, SOL_SOCKET, SO_RCVTIMEO,(char*)&option, sizeof(option));
		setsockopt(con->sock, SOL_SOCKET, SO_SNDTIMEO,(char*)&option, sizeof(option));
		//start receiving informations
		if((con->event = WSACreateEvent())==(HANDLE)-1)
		{
			con->ReportError("Could not create event!",true);
			goto End;
		}
		if(WSAEventSelect(con->sock,con->event,FD_READ | FD_CLOSE)==SOCKET_ERROR)
		{
			con->ReportError("Could not create event!",true);
			goto End;
		}


		con->SetStatus("Autoryzacja...");
		if(con->direction == edIncoming)
		{
			if(!(read = con->RecvEx(8,120000)))
			{
				con->ReportError("Bd odbierania!",true);
				goto End;
			}
			//verify
			int r_uin = ((p2p_init*)con->buffer)->remote_uin;
			int l_uin = ((p2p_init*)con->buffer)->local_uin; // nasz uin...

			if (r_uin == 0x0 && l_uin == 0x0) { // inne sieci
				if(!con->RecvEx(4,1))
					throw eKT("Bd odbierania!",true);
				int flags = *(int*)con->buffer;
				con->net = flags & 0xFF;
				if (((flags & 0xFFFFFF00) != 0) || (con->net != kLAN::net))
					throw eKT("Nieprawidowy format!",true);
				for (int i = 0; i < 2; i++) { // czytamy dwa razy, eby wczyta oba UIDy, ale na razie omijamy drugi...
					if(!con->RecvEx(1,1))
						throw eKT("Bd odbierania!",true);
					size_t c = con->buffer[0];
					if (!c || c > 127)
						throw eKT("Nieprawidowy format!",true);
					if(!con->RecvEx(c,1))
						throw eKT("Bd odbierania UIDa!",true);
					if (i == 0) { // kopiujemy tylko pole "OD"
						memcpy(tbuff , con->buffer , c);
						tbuff[c] = 0;
					}
				}
			} else { // GG
				con->net = NET_GG;
				if(l_uin != GGuin || r_uin == l_uin)
				{
					con->ReportError("Ptla?",false);
					goto End;
				}
				_snprintf(tbuff,256,"%u",r_uin);
			}
			con->cntID = ICMessage(IMC_CNT_FIND, con->net, (int)tbuff);
			if(con->cntID == -1)
			{
				IMLOG("Klient spoza ksiki - %s" , tbuff);
				con->ReportError("Klient spoza ksiki!",false);
				goto End;
			}
			if(!con->SendEx("UDAG",4))
			{
				con->ReportError("Bd wysyania!",true);
				goto End;
			}
		}
		else // Outgoing:
		{
			con->cntID = ICMessage(IMC_CNT_FIND, con->net, (int)con->UID.c_str());
			if (con->net == NET_GG) {
				((p2p_init*)buffer)->remote_uin = GGuin;
				((p2p_init*)buffer)->local_uin = atoi(con->UID.c_str());
				if(!con->SendEx(buffer,8))
					throw eKT("Bd wysyania!",true);
			} else { // inne sieci
				memset(buffer , 0 , 12);
				buffer[8] = con->net & 0xFF;
				if(!con->SendEx(buffer,12))
					throw eKT("Bd wysyania!",true);
				char * myUID = (char*)IMessage(IM_GET_UID , con->net , IMT_ALL);
				if (!myUID)
					throw eKT("Bd wewntrzny!",true);
				size_t len = strlen(myUID);
				if(!con->SendEx(&len,1))
					throw eKT("Bd wysyania!",true);
				if(!con->SendEx(myUID,strlen(myUID)))
					throw eKT("Bd wysyania!",true);
				len = con->UID.size();
				if(!con->SendEx(&len,1))
					throw eKT("Bd wysyania!",true);
				if(!con->SendEx((void*)con->UID.c_str(),con->UID.size()))
					throw eKT("Bd wysyania!",true);
			}

			if(!(read = con->RecvEx(4,120000)) || memcmp(buffer,"UDAG",4)!=0)
			{
				con->ReportError("Bd Autoryzacji!",true);
				goto End;
			}
		}
		
		if(con->direction == edIncoming)
		{
			con->SetStatus ("Czekam na kod akcji...");
			if(!con->RecvEx(4,1))
			{
				con->ReportError("Bd odbierania!",true);
				goto End;
			}

			if(*((DWORD*)con->buffer) == 0x0002)//catch file
			{
				con->operation = eoReceiving;
				con->Stage(esConnected);
				con->ReceiveFile();
			}
			else if(*((DWORD*)con->buffer) == 0x0003)
			{
				con->operation = eoSending;

				// Jeeli Incoming chce, ebymy co wysali, znaczy, e sami sobie
				// tego yczylimy. Szukamy wic odpowiedniego slota na licie...
				EnterCriticalSection(&CSslots);
				for (slots_it i = slots.begin(); i != slots.end(); i++) {
					if ((*i)->cntID == con->cntID && (*i)->stage == esWaiting) {
						// Skoro cntID si zgadza i do tego slot czeka, 
						// uznajemy go za typ request i dopinamy do aktualnego poczenia
						sConnection_request * req = static_cast<sConnection_request*>(*i);
						if (!req) {
							con->ReportError("Bd wewntrzny!!!",true);
							LeaveCriticalSection(&CSslots);
							goto End;
						}
						// Przypisujemy je do siebie
						*con = *req;

						break; // koniecznie przerywamy, bo req wanie dokona sepuku i namiesza na licie
					}
				}
				LeaveCriticalSection(&CSslots);
				con->Stage(esConnected);
				con->SendFile();
			}
			else
			{
				con->ReportError("Bdny kod akcji!",true);
				goto End;
			}
		}
		else // Outgoing:
		{
			con->SetStatus("Wysyam kod akcji...");
			con->Stage(esConnected);
			if(con->operation == eoSending)
			{
				code = 0x0002;
				if(!con->SendEx(&code,4))
				{
					con->ReportError("Bd wysyania!",true);
					goto End;
				}
				con->SendFile();
			}
			else if(con->operation == eoReceiving)
			{
				code = 0x0003;
				if(!con->SendEx(&code,4))
				{
					con->ReportError("Bd wysyania!",true);
					goto End;
				}
				con->ReceiveFile();
			}
			else
			{
				con->ReportError("Bdny kod akcji!",true);
				goto End;
			}
		}
	} catch (eKT e) {
		con->ReportError(e.msg,e.severe);
	} catch (...) {
		con->ReportError("Wystpi nieznany bd!",true);
	}
End:
	delete con;
	return 0;
}


// -------------------------------------------
//
//

void sConnection_request::Start() {
	MakeWindow();
	Stage(esWaiting);
	SetStatus("Oczekuj na poczenie zwrotne...");


	gg_session* ggs = (gg_session*)IMessage(IM_GG_GETSESSION,NET_GG);
	if(ggs)
	{
		CHAR buffer[128] = {0};
		gg_header* pck = (gg_header*)buffer;
		gg_send_msg *msg = (gg_send_msg*)((char*)(buffer + sizeof(gg_header)));

		pck->type = 0x000B; //GG_MSG
		pck->length = 13;//tcp
		msg->seq = 0;
		msg->msgclass = 0x0010; //GG_CLASS_CTCP
		msg->recipient = GETCNTI(this->cntID , CNT_UID);
		buffer[8 + 12]=2;// nie wiem :-)
		gg_write(ggs,buffer,13 + 8);
		IMLOG("Wysaem danie poczenia zwrotnego...");
		IMessage(IM_GG_RELEASESESSION,NET_GG);
	}

}
sConnection_request::~sConnection_request() {
	Stage(esFinished);
//	DoNext();
}



// -------------------------------------------
//
//


VOID sConnection::ReportError(string str,bool fatal)
{
	Stage(esFailure);
	if (this->flags.canceled) return;
	FLASHWINFO fi = {0};
	fi.cbSize =  sizeof(fi);
	fi.uCount = 2;
	fi.hwnd = this->wnd;
	fi.dwFlags = FLASHW_TRAY | FLASHW_CAPTION;
	
	char emsg[32];
	DWORD ec = WSAGetLastError();

	if(ec == 10054)
	{
		SetDlgItemText(this->wnd,IDC_STATUS,"Poczenie zostao anulowane przez drug stron!");
	}
	else
	{
		sprintf(emsg,"WSAE: %u",ec);
		SetDlgItemText(this->wnd,IDC_STATUS,emsg);
	}
	SetDlgItemText(this->wnd,IDC_INFO,str.c_str());
	KTMessage("kTransfer - bd\n" + str , false , KNotify::sIMessage_notify::tError);
	FlashWindowEx(&fi);
}

