/*
Oryginalny kod: Piotr Pawluczuk / piopawlu.prv.pl
Zmiany od wersji 1.1: Rafa Lindemann / www.stamina.eu.org
*/

#include "stdafx.h"

/////////////////////////////////////
const static char* kTransfer_class = "ggprx_class_pp\0";
list <struct sConnection_base *> slots;
int incomingCount = 0;
int connectionsCount = 0;
bool stopRequested = false;
CRITICAL_SECTION CSslots;
/////////////////////////////////////

HANDLE	hMainThread = NULL;
HANDLE  hStopThread = NULL;
UINT	u_main_th = NULL;


/////////////////////////////////////
SOCKET s_main = 0;
SOCKET s_accept = 0;

#define ISVALID(s) (s!=INVALID_SOCKET && s!=0)
#define CLOSESOCK(s) {closesocket(s);s = INVALID_SOCKET;}
#define OK(o) ((o) != SOCKET_ERROR)
#define REPORT(str,fatal){ReportError(str,fatal);goto Error;}

///////////////////////////////
sockaddr_in cli = {0};
int cli_size = sizeof(cli);

// -------------------------------------------------------------------

eKTWSA::eKTWSA(string msg , bool severe):eKT(msg,severe) {
	char buff [10];
	msg += "\nWSA: ["+string(itoa(WSAGetLastError(),buff,10)) + WSAstrerror(WSAGetLastError());
	IMLOG("Error - %s" , msg.c_str());
}


void SocketEvent(HANDLE hSocketEvent) {
	WSANETWORKEVENTS nev = {0};
	WSAEnumNetworkEvents(s_main,hSocketEvent,&nev);
	if(nev.lNetworkEvents & FD_ACCEPT) {
		s_accept = accept(s_main,(sockaddr*)&cli,&cli_size);
		if(s_accept!=INVALID_SOCKET)
		{
			if (incomingCount > GETINT(Cfg::slotsMax)) {
				IMLOG("Too many incoming connections!");
				return;
			}
			sConnection* conn = new sConnection();
			if(!conn)
			{
				throw eKT("Insufficient Memory!",true);
			}
			conn->sock = s_accept;
			conn->direction = sConnection_base::edIncoming;
			conn->address = cli.sin_addr.s_addr;
			conn->port = ntohs(cli.sin_port);
			conn->Start();
		}
		else
		{
			ReportError("Could not accept connection!",false);
		}
	}
}

void ProcessWindowsMessages() {
	MSG msg;
	if (PeekMessage(&msg, NULL, 0, 0 , PM_REMOVE) && msg.hwnd) {
		// Mamy tylko dialogi...
		HWND fhWnd = GetForegroundWindow();
		if (!fhWnd || !IsDialogMessage(fhWnd,&msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
}

int usedPort = 0;
// Gwny wtek
unsigned __stdcall ListenThread(void* param)
{
	stopRequested = false;
	sockaddr_in lst = {0};
	IMLOG("Wtek gwny uruchomiony");
	lst.sin_addr.s_addr = INADDR_ANY;
	lst.sin_family = AF_INET;
	string portVal = GETSTR(Cfg::localPort);
	int minPort = atoi(portVal.c_str());
	int maxPort = atoi(portVal.substr(portVal.find('-')+1).c_str());

	WSAEVENT hSocketEvent = WSACreateEvent();
	hStopThread = CreateEvent(0 , true , false , "Local\\hStopThread");
	ResetEvent(hStopThread);
	if (!hStopThread) throw eKT("Could not create stop event!");
	try {
		s_main = WSASocket(AF_INET,SOCK_STREAM,IPPROTO_IP,NULL,NULL,NULL);
		if(!ISVALID(s_main)) throw eKTWSA("Could not create socket!",true);
		int tryPort = minPort - 1;
		do {
			if (tryPort == maxPort) 
				throw eKT("Could not bind socket!",true);
			lst.sin_port = htons(++tryPort);
			IMLOG("Binding to: %d from (%d - %d)" , tryPort , minPort , maxPort);
		} while (!OK(bind(s_main,(const sockaddr*)&lst,sizeof(lst))));
		usedPort = tryPort;
		if(!OK(listen(s_main,SOMAXCONN))) throw eKTWSA("Could not set SOL_LISTEN!",true);

		if(!OK(WSAEventSelect(s_main, hSocketEvent ,FD_ACCEPT))) throw eKTWSA("Could not set SOL_LISTEN!",true);
		HANDLE events [] = {hStopThread , hSocketEvent};
		unsigned int eventsCount = 2;
		while(1)
		{
	        int ret=MsgWaitForMultipleObjectsEx(eventsCount , events , INFINITE , QS_ALLINPUT | QS_ALLPOSTMESSAGE , MWMO_ALERTABLE | MWMO_INPUTAVAILABLE);
			if (ret - WAIT_OBJECT_0 == 0) { // StopThread
				IMLOG("StopSignal");
				break;
			} else if (ret - WAIT_OBJECT_0 == 1) { //SocketEvent
				SocketEvent(hSocketEvent);
			} else if (ret == WAIT_OBJECT_0 + eventsCount) {
//				GetMessage( &msg, NULL, 0U, 0U );
				ProcessWindowsMessages();
			} else if (ret == WAIT_IO_COMPLETION) {
				IMLOG("APCProc");
			} else {
				IMLOG("Wait zwrci co dziwnego: %d" , ret);
				break;
			}
		}
	} catch (eKTWSA e) {
		KTMessage(e.msg , e.severe , 1);
	} catch (eKT e) {
		KTMessage(e.msg , e.severe , 1);
	}
	IMLOG("Zamykam wszystkie aktywne poczenia:");

	// Zamykamy sloty... 
	// Po anulowaniu kadego przegldamy list od nowa, bo mogo si co pozmienia!
	EnterCriticalSection(&CSslots);
	bool allCanceled = false;
	do {
		allCanceled = true;
		for (slots_it i = slots.begin(); i != slots.end(); i++) {
			IMLOG("  -> slot filename=%s , stage=%d , direction=%d , operation=%d" , (*i)->filename.c_str() , (*i)->stage , (*i)->direction , (*i)->operation);
			if (!(*i)->IsClosing()) {
				(*i)->flags.closed = 1;
				(*i)->Cancel();
				allCanceled = false;
				break;
			}
		}
		SleepEx(0 , TRUE);
		Ctrl->WMProcess();
	} while (!allCanceled);
	LeaveCriticalSection(&CSslots);
	while (slots.size()) {
        if (MsgWaitForMultipleObjectsEx(0 , 0 , 10 , QS_ALLINPUT | QS_ALLPOSTMESSAGE , MWMO_ALERTABLE | MWMO_INPUTAVAILABLE) == WAIT_OBJECT_0)
			ProcessWindowsMessages();
	}
	WSACloseEvent(hSocketEvent);
	CloseHandle(hStopThread);
	hStopThread = 0;
	usedPort = -1;
	IMLOG("Wtek gwny zatrzymany");
	if (!stopRequested) {
		// Wrzucamy informacj o podejrzanym zerwaniu poczenia...
		KTMessage("kTransfer przesta przyjmowa poczenia!" , false , KNotify::sIMessage_notify::tError);
	}
	UpdateKTStatus();
	return 0;
}

////////////////////////////////////////////
//Client Thread
////////////////////////////////////////////


////////////////////////////////////////////
//Miscellaneous
////////////////////////////////////////////


VOID ReportError(const char* str,bool fatal)
{
	char emsg[256];
	DWORD ec = WSAGetLastError();
	sprintf(emsg,"%sError: %s\nWSAE: %u",(fatal)?"Fatal ":"",str,ec);
	KTMessage(emsg , fatal);
}


BOOL CleanUp()
{
	if(hMainThread)CloseHandle(hMainThread);

	hMainThread = NULL;
	u_main_th = 0;

	if(ISVALID(s_main))CLOSESOCK(s_main);
	if(ISVALID(s_accept))CLOSESOCK(s_accept);

	return TRUE;
}

BOOL StartListen()
{
	static bool starting = false;

	if(starting || hMainThread != NULL)return TRUE;
	starting = true;
	usedPort = -1;
	hMainThread = (HANDLE)Ctrl->BeginThread(NULL,NULL,(OLD_THREAD)ListenThread,NULL,NULL,&u_main_th);
	// Czekamy a zostanie ustalony port
	while (usedPort == -1) {
		SleepEx(25 , TRUE);
		Ctrl->WMProcess();
	}
	starting = false;
	UpdateKTStatus();
	if(!hMainThread)return FALSE;
	return TRUE;
}
int StopListen(bool canWork)
{
	static bool working = false;
	if(hMainThread == NULL)return TRUE;
	if (working) {
		while (working) { // skoro ju jest zamykany... POCZEKAJMY!
			Ctrl->Sleep(25);
		}
		return TRUE;
	}
	working = true;
	IMLOG("- Zatrzymuj na yczenie, canWork=%d" , canWork);
	if (slots.size()) {
		if (canWork) {
			if (MessageBox(0 , "S jeszcze aktywne poczenia. Czy na pewno chcesz zatrzyma ich prac?" , "kTransfer" , MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2) == IDNO) {
				working = false;
				IMLOG("- Przerwane");
				return -1;
			}
		}/* else {
			if (MessageBox(0 , "S jeszcze aktywne poczenia. Czy chcesz poczeka a zostan one ukoczone?" , "kTransfer" , MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON1) == IDYES) {
				while (slots.size()) {
					if (MsgWaitForMultipleObjectsEx(0 , 0 , 250 , QS_ALLINPUT | QS_ALLPOSTMESSAGE , MWMO_ALERTABLE | MWMO_INPUTAVAILABLE) == WAIT_OBJECT_0)
						ProcessWindowsMessages();
				}
			}
		}*/

	}
	stopRequested = true;
	if(hStopThread) 
		SetEvent(hStopThread);
	IMLOG("Czekam na zakoczenie gwnego wtku.");
	while (1) {
		if (WaitForSingleObjectEx(hMainThread , 10 , !Ctrl->QuickShutdown() )!=WAIT_TIMEOUT) 
			break;
		Ctrl->WMProcess();
	}
	usedPort = -1;
	//TerminateThread(h_main_th,0);
	CleanUp();
	working = false;
	UpdateKTStatus();
	return TRUE;
}

BOOL IsRunning()
{
	return (hMainThread != NULL);
}
BOOL ThreadInit()
{
   return TRUE;
}
BOOL ThreadDeInit()
{
	return TRUE;
}