#ifndef _kThreads_H_
#define _kThreads_H_

typedef list <struct sConnection_base *> slots_t;
extern slots_t slots;
typedef list <struct sConnection_base *>::iterator slots_it;
extern int usedPort;
extern int incomingCount;
extern int connectionsCount;
extern CRITICAL_SECTION CSslots;
extern SOCKET s_main;
extern bool stopRequested;

typedef UINT (__stdcall *OLD_THREAD)(void*);
/*
#define FLAG_INCOMING 1
#define FLAG_OUTGOING 2
#define FLAG_RECEIVING 4
#define FLAG_SENDING 8
#define FLAG_CALLBACK 16
#define FLAG_ACTIVE 32
#define FLAG_NOCANCEL 64
*/
// exception
struct eKT {
	string msg;
	bool severe;
	eKT(string msg , bool severe = false):msg(msg),severe(severe) {}
};
struct eKTWSA:public eKT {
	eKTWSA(string msg , bool severe = false);
};

struct sConnection_base {
	int		cntID; // ID kontaktu...
	sConnection_base * next; // Nastpny do wywoania...
	HWND	wnd;
	string	filename;

	enum enStage {
		esNone,
		esCreated,
		esWaiting, // dla _request
		esConnecting,
		esConnected, // poczony I PO autoryzacji
		esRefresh,
		esNegotiating,
		esOperating,
		esFailure,
		esSuccess, // caa operacja zostaa wykonana
		esCancel, // w trakcie anulowywania...
		esFinished
	} stage; 
	enum enDirection {
		edUnknown,
		edIncoming,
		edOutgoing
	} direction;
	enum enOperation {
		eoUnknown,
		eoReceiving,
		eoSending
	} operation;

	struct {
		int callback:1;
		int active:1;
		int nocancel:1;
		int canceled:1;
		int success:1;
		int closed:1;
		int fromRequest:1; 
	} flags;


	virtual void Start(){};
	virtual void Stage(enStage stage); // Wywoane w odpowiednim stadium poczenia
	void SetStatus(const char * msg);
	void MakeWindow();
	virtual void ShowWindow(bool foreground);
	virtual void SetWindowTitle();
	static VOID CALLBACK sConnection_base::MakeWindowAPC(sConnection_base * c);
	static VOID CALLBACK sConnection_base::DestroyWindowAPC(sConnection_base * c);
	void DoNext();
	virtual void Cancel();
	sConnection_base();
	virtual ~sConnection_base();
	static BOOL WINAPI sConnection_base::DlgProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
	bool IsClosing() {return flags.canceled || flags.closed || stage == esFinished || stage == esCancel; }
	virtual void Retry() {}
	int CountNext();
	sConnection_base & operator = (sConnection_base & old); 

};

// Poczenie wymagajce wysania komunikatu GG z prob o poczenie zwrotne
struct sConnection_request: public sConnection_base {
	void Start();
	void Stage(enStage stage); // Wywoane w odpowiednim stadium poczenia
	sConnection_request(struct sConnection & req);
	sConnection_request() {}
	~sConnection_request();
};


struct sConnection: public sConnection_base
{


	HANDLE	thread;
	HANDLE  event;
	HANDLE	file;
	HANDLE  cancelEvent;
	UINT	threadID;
	SOCKET	sock;
	UINT	protocol;
	BYTE	buffer[16384 + 2];

	DWORD	options;
	DWORD	done;
	DWORD   transfered;
	DWORD	total;


	int msgID; /* ID wiadomoci, ktra ma otworzy to okienko lub 0
				*/
	int notifyID;

	DWORD	address;
	USHORT	port;
	std::string UID;
	int		net;
	time_t  startTime;
	int lastUpdate;


	sConnection();
	sConnection(sConnection & req);
	~sConnection();
	sConnection & operator = (sConnection_request & req); 
	sConnection & operator = (sConnection & req); 
	void Retry();
	void Start();
	void Stage(enStage stage); // Wywoane w odpowiednim stadium poczenia
	void DropToHistory(enStage lastStage);
	unsigned int RecvEx(unsigned int toread,int reqall);
	void UpdateInfo(double transfer = -1);
	void ShowWindow(bool foreground);

	long SendEx(void* buffer,long length);
	long SendFile();
	long ReceiveFile();
	static unsigned __stdcall ClientThread(sConnection* con);
	VOID ReportError(string str,bool fatal);
	void Cancel();
	void SetWindowTitle();

	void init();

};

// Zmienne
//
//


// Funkcje
//
//

BOOL StartListen();
int StopListen(bool canWork);
BOOL ThreadInit();
BOOL ThreadDeInit();
BOOL IsRunning();

// Threads

LONG RecvEx(sConnection* sc,LONG toread,BOOL reqall);
LONG SendEx(sConnection* sc,void* buffer,long length);

unsigned __stdcall ClientThread(sConnection* con);
BOOL CleanUp();
VOID ReportError(const char* str,bool fatal);
VOID ReportError(const char* str,bool fatal,sConnection* con);
LRESULT WINAPI WndProc(HWND hWnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
BOOL WINAPI DlgProc(HWND hDlg,UINT uMsg,WPARAM wParam,LPARAM lParam);
void SocketEvent();

LONG ReceiveFile(sConnection* sc);

#endif //_kThreads_H_